var EventEmitter = require('events').EventEmitter,
	util = require('util'),
	wcwidth = require('wcwidth.js')({
		nul:     0,
		control: 1
	})

module.exports = function(nc, inputWin) {
	function inputbuffer() {
		EventEmitter.call(this)
		this.buffer = ''
		// difference between trueCursor and inputWin.curx is that trueCursor will NOT be limited by input window width
		// it will be "true" to the input buffer itself, not to the current displayed input...
		this.trueCursor = 0
		// inputOffset is where the current "start" of input actually is along the current string buffer
		this.inputOffset = 0

		// @todo for later, shifting by more than one character at a time when there's too much to handle at once
		// note: this is width-relative (ncurses window cursor depends on width!), not length-relative
		this.shiftTolerance = 4
		this.shiftEdge = 2

		// character prefix for commands - if it doesn't have it at the very beginning, it ain't a command.
		this.cmdPrefix = ':'
		this._cmdPrefix = new RegExp('^' + this.cmdPrefix.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&") + '(\\w+|\\$([0-9a-z]{2}))\\s*')
	}
	util.inherits(inputbuffer, EventEmitter)

	inputbuffer.prototype.inputChar = function(char, charCode, isKey) {
		var self = this
		if(charCode === nc.keys.ESC) {
			this.emit('command:quit')
		} else if(charCode === 18) { // ^r
			this.emit('command:refresh')
		} else if(charCode === nc.keys.BACKSPACE || charCode === 127 /* ASCII DEL */) {
			// silently ignore if we're at the beginning of the string
			if(this.trueCursor > 0 && this.buffer.length > 0) {
				var prev_x = inputWin.curx
				inputWin.delch(0, this.trueCursor - this.inputOffset - 1)
				this.buffer = this.buffer.substring(0, this.trueCursor - 1) + this.buffer.substring(this.trueCursor)
				this.trueCursor--
				inputWin.cursor(0, prev_x - 1)

				if(inputWin.curx <= this.shiftEdge) {
					for(var i = 0, chWidth = this.shiftTolerance; chWidth > 0 && (this.inputOffset - i) > 0; i++)
						chWidth -= this.buffer[this.inputOffset - i].wcwidth

					this.inputOffset -= i
					this.redraw(this.inputOffset, prev_x + i - 1)
				}
				inputWin.refresh()
			}
		} else if (charCode === 23) { // ^w
			// @todo delete word
		} else if (charCode === nc.keys.DEL) {
			var prev_x = inputWin.curx
			if(this.buffer.length > 0 && this.trueCursor < this.buffer.length) {
				inputWin.delch(0, inputWin.curx)
				this.buffer = this.buffer.substring(0, this.trueCursor) + this.buffer.substring(this.trueCursor + 1)
				inputWin.cursor(0, prev_x)
				inputWin.refresh()
			}
		} else if(charCode === nc.keys.LEFT) {
			if(this.trueCursor > 0) {
				var prev_x = inputWin.curx
				inputWin.cursor(0, prev_x - 1)
				this.trueCursor--

				if(inputWin.curx <= this.shiftEdge) {
					for(var i = 0, chWidth = this.shiftTolerance; chWidth > 0 && (this.inputOffset - i) > 0; i++)
						chWidth -= this.buffer[this.inputOffset - i].wcwidth
					this.inputOffset -= i
					this.redraw(this.inputOffset, prev_x + i - 1)
				}
				inputWin.refresh()
			}
		} else if(charCode === nc.keys.RIGHT) {
			if(this.trueCursor < this.buffer.length) {
				var prev_x = inputWin.curx
				inputWin.cursor(0, inputWin.curx + 1)
				this.trueCursor++

				// are we shifting?
				if(inputWin.curx >= inputWin.width - this.shiftEdge) {
					// shift characters - REDRAW!
					for(var i = 1, chWidth = this.shiftTolerance; chWidth > 0; i++)
						chWidth -= this.buffer[this.inputOffset + i].wcwidth

					this.inputOffset += i
					this.redraw(this.inputOffset, prev_x - i)
				}
				inputWin.refresh()
			}
		} else if(charCode === nc.keys.UP) {
			// @todo entry history
		} else if(charCode === nc.keys.DOWN) {
			// @todo entry history
		} else if(charCode === nc.keys.HOME) {
			// ...unusually easy.
			this.trueCursor = 0
			this.inputOffset = 0
			this.redraw(0,0)
			inputWin.refresh()
		} else if(charCode === nc.keys.END) {
			// try to guess at how far we'll need to be...may need to adjust for east-asian character width though for proper display
			this.trueCursor = this.buffer.length
			if(this.buffer.wcwidth > inputWin.width - (this.shiftTolerance + this.shiftEdge)) {
				// we need to figure out line-shifting, ugh
				var maxWidth = inputWin.width - (this.shiftTolerance + this.shiftEdge),
					offset = this.buffer.length - maxWidth,
					displayBuffer = this.buffer.slice(offset),
					curWidth = displayBuffer.wcwidth

				// WILL IT FIT?
				if(curWidth > maxWidth) {
					// need to increase the offset so that total width will decrease, allowing everything to fit
					for(var i = 1, chWidth = maxWidth - curWidth; chWidth > 0; i++)
						chWidth -= this.buffer[this.offset + i].wcwidth
					offset += i
				}
				this.inputOffset = offset
				this.redraw(offset, this.buffer.slice(offset).wcwidth)
			} else {
				// EASY MODE
				this.redraw(0, this.buffer.wcwidth)
			}
			inputWin.refresh()
		} else if(charCode === nc.keys.NEWLINE) {
			if(this.buffer.length > 0) { // shit happens...
				this.commandParser(this.buffer)
				this.buffer = ''
				inputWin.insdelln(-1)
				inputWin.cursor(0,0)
				inputWin.refresh()
				this.trueCursor = 0
			}

		} else if(charCode === nc.keys.NPAGE) {
			this.emit('pageup')
		} else if(charCode === nc.keys.PPAGE) {
			this.emit('pagedown')
		} else {
			if(this.trueCursor === this.buffer.length) {
				var prev_x = inputWin.curx
				inputWin.print(char)
				this.buffer += char
				this.trueCursor++

				// are we shifting?
				if(inputWin.curx >= inputWin.width - this.shiftEdge) {
					// shift characters - REDRAW!
					for(var i = 1, chWidth = this.shiftTolerance; chWidth > 0; i++)
						chWidth -= this.buffer[this.inputOffset + i].wcwidth

					this.inputOffset += i
					this.redraw(this.inputOffset, prev_x - i + 1)
				}
			} else {
				// we're inserting, may need to redraw.
				var prev_x = inputWin.curx
				this.buffer = this.buffer.substring(0, this.trueCursor) + char + this.buffer.substring(this.trueCursor)
				this.trueCursor++
				this.redraw(this.inputOffset, prev_x + 1)
			}
			inputWin.refresh()
		}

		this.updateChars()
	}

	inputbuffer.prototype.updateChars = function() {
		if(this.buffer.match(this._cmdPrefix) !== null || this.buffer.length == 0) {
			this.emit('char:update', false)
		} else {
			this.emit('char:update', '' + (140 - this.buffer.length))
		}
	}

	inputbuffer.prototype.redraw = function(offset, newCursor) {
		if(newCursor === undefined) newCursor = inputWin.curx
		if(offset === undefined) this.inputOffset

		inputWin.insdelln(-1)
		inputWin.cursor(0,0)
		inputWin.print(this.buffer.slice(offset, offset + inputWin.width))
		inputWin.cursor(0, newCursor)
	}

	inputbuffer.prototype.commandParser = function(buffer) {
		// check if command
		var cmd = buffer.match(this._cmdPrefix), args = '', lineId = ''
		if(cmd !== null) {
			args = buffer.slice(cmd[0].length)

			// short form reply, for the lazy
			// ex :$0a my reply here
			if(cmd[1][0] === '$') {
				lineId = cmd[2]
				this.emit('command:reply', lineId.toLowerCase(), args) // format: fn(id, replyMessage)
				return
			}

			switch(cmd[1].toLowerCase()) {
				case 'reply':
				case 'r':
					lineId = args.match(/^\$([0-9a-z]{2})\s+/i)
					if(lineId === null) {
						this.emit('command:error', 'line ID not specified properly')
						return
					}
					var reply = args.slice(lineId[0].length)
					this.emit('command:reply', lineId[1].toLowerCase(), reply) // format: fn(id, replyMessage)
					break

				case 'retweet':
				case 'rt':
					lineId = args.match(/^\$([0-9a-z]{2})\s*/i)
					if(lineId === null) {
						this.emit('command:error', 'line ID not specified properly')
						return
					}
					this.emit('command:retweet', lineId[1].toLowerCase()) // format: fn(id)
					break

				case 'favorite':
				case 'fav':
					lineId = args.match(/^\$([0-9a-z]{2})\s*/i)
					if(lineId === null) {
						this.emit('command:error', 'line ID not specified properly')
						return
					}
					this.emit('command:favorite', lineId[1].toLowerCase()) // format: fn(id)
					break

				case 'unfavorite':
				case 'unfav':
					lineId = args.match(/^\$([0-9a-z]{2})\s*/i)
					if(lineId === null) {
						this.emit('command:error', 'line ID not specified properly')
						return
					}
					this.emit('command:unfavorite', lineId[1].toLowerCase()) // format: fn(id)
					break

				case 'delete':
				case 'del':
					lineId = args.match(/^\$([0-9a-z]{2})\s*/i)
					if(lineId === null) {
						this.emit('command:error', 'line ID not specified properly')
						return
					}
					this.emit('command:delete', lineId[1].toLowerCase()) // format: fn(id)
					break

				case 'link':
					lineId = args.match(/^\$([0-9a-z]{2})\s*/i)
					if(lineId === null) {
						this.emit('command:error', 'line ID not specified properly')
						return
					}
					this.emit('command:link', lineId[1].toLowerCase()) // format: fn(id)
					break

				case 'table':
					this.emit('command:tweet', '\u0028\u256f\u00b0\u0414\u00b0\uff09\u256f\u5f61\u0020\u253b\u2501\u253b')
					break

				// @todo remove
				case 'redraw':
					nc.redraw()
					break

				case 'quit':
				case 'q':
					this.emit('command:quit')
					break

				default:
					this.emit('command:unk')
					break
			}
			return
		} else {
			// @todo 140-char buffer length check?
			this.emit('command:tweet', buffer)
		}
	}

	return new inputbuffer()
}
