var qs = require('querystring'),
	fs = require('fs'),
	util = require('util'),
	nc = require('ncurses'),
	request = require('request'),
	nconf = require('nconf'),
	//hook = require('./hook'),
	scrollbuffer = require('./scrollbuffer'),
	filter = require('./filter'),
	tweetstream = require('./tweetstream'),
	uimessage = require('./uimessage'),
	pkg = require('../package.json')

function rumble(nconf, confFile) {
	// yes, this isn't the best of ideas having the twitter oauth consumer key/secret in plaintext right here
	// however, due to javascript's role as a plaintext langauge (not compiled) and that any attempts to cipher or encrypt them require an easy way
	// for users to decrypt them in order to make the client usable means that it doesn't make sense to attempt full-blown encryption when it can just be easily reversed.
	// instead we'll just to a little bit of obfuscation to make automated scanners choke and move on.
	//
	// don't blame me, blame oauth for a shit design when used with open source products.
	// a better way would have been client certificate authentication with chain-signing...but meh. bitching gets you nowhere with twitter hq.

	this.confFile = confFile
	//this.nconf = nconf
	var options = nconf.get('conf:twitter')

	this.consumer_key = options.consumer_key || new Buffer('4370684b73534f586e567430666e3933474e64695577', 'hex').toString('utf8')
	this.consumer_secret = options.consumer_secret || new Buffer('54364b4658305941477255676f7671696238354562466733476c5977577153336a643055775a696930', 'hex').toString('utf8')
	this.access_token = options.access_token || ''
	this.access_token_secret = options.access_token_secret || ''
	this.selfID = options.self_user_id || 0
}
rumble.prototype.runtime = function(app) {
	// useful references for implementation
	//
	// https://github.com/mscdex/node-ncurses
	// https://github.com/mscdex/node-ncurses/blob/master/examples/irc.client.js
	// https://github.com/rithis/tio
	// https://github.com/rithis/tio/blob/master/ui.js
	var self = this

	if(!fs.existsSync(self.confFile) || self.access_token === '' || self.access_token_secret === '') {
		console.error('Authentication details missing, please run "rumble -a" first')
		process.exit(1)
	}

	process.on('uncaughtException', function (ex) {
		nc.cleanup()
		//hook.emit('client:killed')
		// if we have a stack for this, let's grab it
		if(ex.stack) {
			console.error('Uncaught exception: \n' + ex.stack)
		} else {
			console.error('Uncaught exception: ' + util.inspect(ex))
		}

		process.exit(1)
	})
	process.on('SIGINT', function () {
		nc.cleanup()
		//hook.emit('client:killed')
		console.error('Received SIGINT - terminated')
		process.exit(1)
	})
	process.on('SIGTERM', function() {
		nc.cleanup()
		//hook.emit('client:killed')
		console.error('Received SIGTERM - terminated')
		process.exit(1)
	})

	try {
		// init filter, tweetstream...
		var tfilter = new filter(this.confFile),
			tstream = new tweetstream({
				consumer_key: self.consumer_key,
				consumer_secret: self.consumer_secret,
				access_token: self.access_token,
				access_token_secret: self.access_token_secret
			}, this.selfID, tfilter)

		// handle ncurses UI preparation here
		var stdscr = new nc.Window(),
			mainWin = new nc.Window(stdscr.height - 2, stdscr.width - 6),
			idWin = new nc.Window(mainWin.height, 6),
			inputWin = new nc.Window(1, stdscr.width - 2),
			charWin = new nc.Window(1, 4),
			statusWin = new nc.Window(1, stdscr.width),
			sbuffer = new scrollbuffer(mainWin.height, mainWin.width),
			theme = require('./theme')(nc)
			ibuffer = require('./inputbuffer')(nc, inputWin),
			statusbar = require('./statusbar')

		// prepare mainwin
		mainWin.move(0,6)
		mainWin.idlok(true)
		mainWin.scrollok(true)
		mainWin.setscrreg(0, mainWin.height - 1)
		mainWin.refresh()

		// prepare idwin
		idWin.move(0,0)
		idWin.idlok(true)
		idWin.scrollok(true)
		idWin.setscrreg(0 ,idWin.height)
		idWin.refresh()

		// prepare statuswin & charWin
		statusWin.move(stdscr.height - 1, 0)
		charWin.move(stdscr.height - 1, stdscr.width - charWin.width)

		// prepare inputwin
		stdscr.cursor(stdscr.height - 2, 0)
		//stdscr.print('> ')
		stdscr.print('\u266a ')
		stdscr.cursor(0, 0)
		stdscr.refresh()
		inputWin.move(stdscr.height - 2, 2)
		inputWin.top()
		inputWin.cursor(0,0)
		inputWin.refresh()

		// hook up inputbuffer to inputWin key events
		inputWin.on('inputChar', function() {
			ibuffer.inputChar.apply(ibuffer, arguments)
		})

		var statusReset = null
		statusbar.on('update', function(message, style) {
			statusWin.insdelln(-1)
			statusWin.cursor(0,0)
			theme.apply(statusWin, style, message + new Array(statusWin.width - message.length + 1).join(' '))
			statusWin.refresh()
		})
		statusbar.on('showchars', function() {
			charWin.show()
			charWin.refresh()
			inputWin.top()
			inputWin.refresh()
		})
		statusbar.on('hidechars', function() {
			charWin.hide()
			inputWin.top()
		})

		// xorg sigwinch handler
		// non-functional due to node-ncurses bug
		process.on('SIGWINCH', function() {
			//hook.emit('client:resize')
			stdscr.resize(nc.lines, nc.cols)
			//win.resize(process.stdout.rows, process.stdout.columns)
			mainWin.refresh()
			idWin.refresh()
			inputWin.refresh()
			statusWin.refresh()
		})

		ibuffer.on('command:quit', function() {
			tstream.close()
			stdscr.close() // closing stdscr should close all child windows
			//hook.emit('client:closed')
			process.exit(0)
		})
		ibuffer.on('command:refresh', function() {
			idWin.refresh()
			mainWin.refresh()
		})
		ibuffer.on('command:reply', function(lineId, message) {
			line = sbuffer.getLine(scrollbuffer.stringIdToIntId(lineId))
			if(line !== null && line.isTweet) {
				message = line.mentions.slice().reduce(function(ret, mention) {
					if(tstream.me !== null && mention === tstream.me.screen_name) {
						return ret
					} else {
						return ret + ' @' + mention
					}
				}, '@' + line.author) + ' ' + message
				tstream.newTweet(message, line.id)
			} else {
				statusbar.flash(util.format('[error] invalid line %s', lineId), 2, 'STATUSBAR_WARN')
			}
		})

		ibuffer.on('command:retweet', function(lineId) {
			line = sbuffer.getLine(scrollbuffer.stringIdToIntId(lineId))
			if(line !== null && line.isTweet) {
				tstream.newRetweet(line.id)
			} else {
				statusbar.flash(util.format('[error] invalid line %s', lineId), 2, 'STATUSBAR_WARN')
			}
		})

		ibuffer.on('command:favorite', function(lineId) {
			line = sbuffer.getLine(scrollbuffer.stringIdToIntId(lineId))
			if(line !== null && line.isTweet) {
				tstream.favTweet(line.id)
			} else {
				statusbar.flash(util.format('[error] invalid line %s', lineId), 2, 'STATUSBAR_WARN')
			}
		})

		ibuffer.on('command:unfavorite', function(lineId) {
			line = sbuffer.getLine(scrollbuffer.stringIdToIntId(lineId))
			if(line !== null && line.isTweet) {
				tstream.unfavTweet(line.id)
			} else {
				statusbar.flash(util.format('[error] invalid line %s', lineId), 2, 'STATUSBAR_WARN')
			}
		})

		ibuffer.on('command:delete', function(lineId) {
			line = sbuffer.getLine(scrollbuffer.stringIdToIntId(lineId))
			if(line !== null && line.isTweet && line.isMine) {
				tstream.deleteTweet(line.id)
			} else if(line !== null && line.isTweet && !line.isMine) {
				statusbar.flash(util.format('[error] cannot delete tweet %s, we do not own it', lineId), 2, 'STATUSBAR_WARN')
			} else {
				statusbar.flash(util.format('[error] invalid line %s', lineId), 2, 'STATUSBAR_WARN')
			}
		})

		ibuffer.on('command:link', function(lineId) {
			line = sbuffer.getLine(scrollbuffer.stringIdToIntId(lineId))
			if(line !== null && line.isTweet) {
				sbuffer.insert(new uimessage(util.format('[$%s] <https://twitter.com/%s/status/%s>', lineId, line.author, line.id), 'normal'))
			} else {
				statusbar.flash(util.format('[error] invalid line %s', lineId), 2, 'STATUSBAR_WARN')
			}
		})

		ibuffer.on('command:tweet', function(message) {
			tstream.newTweet(message)
		})
		ibuffer.on('command:error', function(err) {
			statusbar.flash(err, 3, 'STATUSBAR_WARN')
		})

		ibuffer.on('char:update', function(chars) {
			if(chars === undefined || chars === false) {
				chars = '    '
			} else {
				chars = '' + new Array(charWin.width - chars.length + 1).join(' ') + chars
				chars = chars.slice(0,4)
			}
			charWin.deleteln()
			charWin.cursor(0,0)
			theme.apply(charWin, 'STATUSBAR', chars)
			charWin.refresh()
		})

		// on new scrollbuffer line inserts, update and display them.
		sbuffer.on('update', function(lineId, height) {
			var line = sbuffer.getLine(lineId)
			//hook.emit('client:ui:newline', lineId, height)
			mainWin.cursor(mainWin.height - 1, 0)
			idWin.cursor(idWin.height - 1, 0)
			mainWin.scroll(1)
			idWin.scroll(1)

			// handle "theming" of tweet parts and pieces
			line.text().forEach(function(chunk) {
				switch(chunk.type) {
					case 'url':
						theme.apply(mainWin, 'TWEET_URL', chunk.content)
						break
					case 'hashtag':
						theme.apply(mainWin, 'TWEET_HASHTAG', chunk.content)
						break
					case 'mention':
						theme.apply(mainWin, 'TWEET_MENTION', chunk.content)
						break
					case 'date':
						theme.apply(mainWin, 'TWEET_DATE', chunk.content.replace(chunk.extra[0], chunk.extra[1]))
						break
					case 'extra':
						theme.apply(mainWin, 'TWEET_EXTRA', chunk.content)
						break
					case 'source':
						theme.apply(mainWin, 'TWEET_SOURCE', chunk.content.replace(chunk.extra[0], chunk.extra[1]))
						break
					case 'author':
						theme.apply(mainWin, 'TWEET_AUTHOR', chunk.content)
						break
					case 'ui_message':
						theme.apply(mainWin, 'UI_MESSAGE', chunk.content)
						break
					case 'text':
					default:
						theme.apply(mainWin, 'TWEET_TEXT', chunk.content)
						break
				}
			})

			var idMsg = '====='
			if(line.isTweet === true)
				idMsg = util.format('[$%s]', line.displayId)

			switch(line.type) {
				case 'highlight':
					theme.apply(idWin, 'IDBAR_HIGHLIGHT', idMsg)
					break
				case 'alert':
					theme.apply(idWin, 'IDBAR_ALERT', idMsg)
					break
				case 'notice':
					theme.apply(idWin, 'IDBAR_NOTICE', idMsg)
					break
				case 'quiet':
					// pass - no entry for idWin
					break
				case 'normal':
				default:
					theme.apply(idWin, 'IDBAR_NORMAL', idMsg)
					break
			}
			if(line.height > 1)
				idWin.scroll((line.type !== 'quiet') ? line.height - 1 : line.height)

			mainWin.refresh()
			idWin.refresh()
		})

		sbuffer.on('redraw', function(lines) {
			// asdf
		})

		// we're "ready"
		//hook.emit('client:ready')
		statusbar.hold('[connecting...]')
		ibuffer.emit('char:update', false)
		tstream.connect(function() {
			sbuffer.insert(new uimessage('rumble, SPIN ON!', 'alert'))
			tstream.getBackfill(20, function(err, tweets) {
				if(err) return // handled by twitter:error emit
				tweets.forEach(function(tweetObj) {
					sbuffer.insert(tweetObj)
				})
			})
		})
		tstream.on('stream:connect', function() {
			statusbar.flash('[connected!]', 1)
		})

		tstream.on('stream:tweet', function(tweetObj) {
			sbuffer.insert(tweetObj)
		})
		tstream.on('stream:filteredtweet', function(tweetObj, filterReason) {
			console.error(filterReason + ' , id: ' + tweetObj.id) // debug
		})

		tstream.on('stream:error', function(err) {
			throw err // ohsh*t
		})
		tstream.on('twitter:error', function(err) {
			if(err.twitterReply) {
				err = JSON.parse(err.twitterReply).errors.reduce(function(ret, errObj) {
					return ret.concat(errObj.message)
				}, []).join(', ')
			}
			statusbar.flash(util.format('[error] %s', err), 2, 'STATUSBAR_ERR')
		})

		tstream.on('stream:reconnect', function(err) {
			statusbar.hold('[reconnecting...]', 'STATUSBAR_WARN')
		})
		tstream.on('stream:disconnect', function(err) {
			statusbar.hold(util.format('[disconnected: %s]', err.reason), 'STATUSBAR_WARN')
		})
	} catch(err) {
		nc.cleanup()
		console.error((err.stack) ? err.stack : err)
		console.log((err.stack) ? err.stack : err)
		process.exit(1)
	}
}
rumble.prototype.authorize = function(app) {
	var self = this
	try {
		var oauth = {
			callback: 'oob',
			consumer_key: self.consumer_key,
			consumer_secret: self.consumer_secret,
		}
		console.log('requesting authorization from twitter, please wait')
		request.post({url: 'https://api.twitter.com/oauth/request_token', oauth:oauth}, function (err, res, body) {
			if(err) throw new Error('Authorization error: ' + err)

			var access_token = qs.parse(body)
			console.log('please open <https://api.twitter.com/oauth/authorize?oauth_token=%s> in your browser to continue, then enter the pin provided below', access_token.oauth_token)
			app.prompt('PIN: ', function(pin) {
				oauth = {
					consumer_key: self.consumer_key,
					consumer_secret: self.consumer_secret,
					token: access_token.oauth_token,
					token_secret: access_token.oauth_token_secret,
					verifier: pin,
				}
				console.log('obtaining access tokens from twitter, please wait')
				request.post({url: 'https://api.twitter.com/oauth/access_token', oauth:oauth}, function(err, res, body) {
					if(err) throw new Error('Authorization error: ' + err)

					var perm_token = qs.parse(body), details = {
						conf:{
							twitter:{
								consumer_key: self.consumer_key,
								consumer_secret: self.consumer_secret,
								access_token: perm_token.oauth_token,
								access_token_secret: perm_token.oauth_token_secret,
								self_user_id: perm_token.user_id
							}
						}
					}

					fs.writeFileSync(self.confFile, JSON.stringify(details, null, 2), 'utf8')
					console.log('Authorization complete')

					// we'll terminate here and instruct the user to re-run normally so we can re-init.
					console.log()
					process.exit(0)
				})
			})
		})
	} catch(err) {
		console.error(err)
		process.exit(1)
	}
}

module.exports = rumble
