var EventEmitter = require('events').EventEmitter,
	util = require('util')

// scrollbuffer, son of EventEmitter, son of Thor. let us hear your mighty roar.
function scrollbuffer(height, width) {
	args = Array.prototype.slice.call(arguments)
	if(width !== undefined) args.shift()
	this.maxSize = 36 * 36 // [0-9a-z]{2} or [00] to [zz] and all combinations in between - results in a buffer size of 1296
	this.lines = {}
	this.dataIdToLine = []
	this.height = height
	this.width = width
	this.visible = {}
	this.autoscroll = true
	this.current = null
	EventEmitter.apply(this, args)
}
util.inherits(scrollbuffer, EventEmitter)

scrollbuffer.prototype.showRange = function(start, fn) {
	if(start === null) start = this.current

	// disable autoscroll if we're not starting at the latest entry
	this.autoscroll = (start === this.current) ? true : false

	var ret = this.getRange(start, this.height, true)
	fn(ret)

	// rebuild visible range
	this.visible = {}
	var size = this.height, shown = 0
	ret.forEach(function(line) {
		size -= line.height
		if(size > 0) {
			shown = line.height
		} else {
			shown = size + line.height
		}
		this.visible[line.displayId] = [line.height, shown]
	})
}

scrollbuffer.prototype.getRange = function(start, size, stopAtCurrent) {
	var ret = [], i, retSize = 0, missed = 0, direction = size > 0 ? -1 : 1 // if size is negative (size * -1), we're doing a backwards range, old to new.
	size *= direction // negative size will be forced positive this way

	if(start > this.lines.length) { // wat
		return [] // bail out now...wtf.
	}

	// note: retSize does NOT refer to length of ret array
	// it refers to the sum of the "line heights" (vertical height, amount of newlines in line.text + 1)!
	for(i = start; retSize < size; i += direction) {
		// id looparound
		if(i === -1) {
			i = 1295
		} else if(i === 1296) {
			i = 0
		}

		if(i === this.current && stopAtCurrent === true) {
			if(direction < 0) // if we're going up the stack (old to new) we want to include current in the results
				ret.push(this.lines[i])
			return ret
		} else if(this.lines[i] === undefined) {
			// bail out, not enough data available to continue.
			// this is an indicator of too high of a "start" point, or too long of a "size",
			// to the point that, on a clean start, we're asking for data that just *is not there* yet.
			// this will imply that the buffer has not begun to roll over on itself as well.
			return ret
		} else if (this.lines[i] === null) {
			// null entries actually represent deleted entries.  We continue on and keep trying to gather data anyways
			continue
		} else {
			ret.push(this.lines[i])
			retSize += this.lines[i].height
		}
	}

	return ret
}

scrollbuffer.prototype.insert = function(line) {
	// get line id - if it's null, just assume 0
	var lineId = (this.current !== null) ? this.nextId(true) : this.current = 0
	line.displayId = scrollbuffer.intIdToStringId(lineId)
	this.lines[lineId] = line.build(this.width)
	this.dataIdToLine[line.id] = lineId // so we can delete based on tweet id...

	if(this.autoscroll)
		this.emit('update', lineId, line.height) // ignore if we're not autoscrolling

	return this
}

scrollbuffer.prototype.updateVisible = function(newLine) {
	// we have a set height already defined
	// let's see if anything is bumped out and can be considered no longer visible on-screen now.
	// we'll start, as always, by reversing the array for ourselves to process. Should be faster that way.
	Object.keys(this.visible).reverse().forEach(function(line) {
		// asdf
		// @todo finish
	})
}

scrollbuffer.prototype.delLineByDisplayId = function(line) {
	var self = this
	if(line.displayId !== undefined && line.displayId !== null)
		line = line.displayId

	if(this.lines[line] !== undefined && this.lines[line] !== null) {
		delete this.dataIdToLine[this.lines[line].id]
		this.lines[line] = null
	}

	// signal redraw of scrollbuffer if line was still marked as visible
	if(this.visible.hasOwnProperty(line))
		self.emit('redraw')

	return this
}

scrollbuffer.prototype.delLineByDataId = function(id) {
	var line = (this.dataIdToLine[id] !== undefined || this.dataIdToLine[id] !== null) ? this.dataIdToLine[id] : null

	if(line !== null)
		this.delLineByDisplayId(line)

	return this
}

scrollbuffer.prototype.getLine = function(line) {
	if(this.lines[line] !== undefined) {
		return this.lines[line]
	} else {
		return null
	}
}

scrollbuffer.prototype.redraw = function(height, width) {
	this.height = height
	this.width = width
	// screen managing code will have to determine topmost "line" id and then regrab the range of ids
	for(var i = 0; i < 1296; i++) {
		if(this.lines[i] === null) {
			continue
		} else if(this.lines[i] === undefined) {
			return
		} else {
			this.lines[i].redraw(width)
		}
	}
}

scrollbuffer.prototype.prevId = function(advance) {
	advance = (advance === true) ? true : false // must be true (not undefined!) in order to 'advance'...backwards
	currentId = this.current - 1
	if(currentId === -1) currentId = 1295
	if(advance) this.current = currentId

	return currentId
}

scrollbuffer.prototype.nextId = function(advance) {
	advance = (advance === true) ? true : false // must be true (not undefined!) in order to 'advance'...backwards

	currentId = this.current + 1
	if(currentId === 1296) currentId = 0
	if(advance) this.current = currentId

	return currentId
}

scrollbuffer.intIdToStringId = function(id) {
	var ids = [Math.floor(id / 36), id % 36]
	ids[0] = (ids[0] > 9) ? String.fromCharCode(ids[0] + 87) : ids[0]
	ids[1] = (ids[1] > 9) ? String.fromCharCode(ids[1] + 87) : ids[1]

	return ids.join('')
}

scrollbuffer.stringIdToIntId = function(ids) {
	var ret = 0
	if(ids.length !== 2)
		return null

	ret += (isNaN(ids[0]) ? ids.charCodeAt(0) - 87 : parseInt(ids[0])) * 36
	ret += isNaN(ids[1]) ? ids.charCodeAt(1) - 87 : parseInt(ids[1])

	return ret
}

module.exports = scrollbuffer
