var cheerio = require('cheerio'),
	dateFormat = require('dateformat'),
	_ = require('underscore'),
	line = require('./line'),
	util = require('util')

function tweet(data) {
	var self = this
	line.call(this, 'normal')

	this.isTweet = true
	this.id = data.id_str
	this.isRetweet = data.retweeted_status !== undefined ? true : false
	this.isReply = data.in_reply_to_user_id !== null ? true : false

	// these properties are initialized here, though unused - we will set them in tweetstream instead.
	this.isMine = false
	this.isMentioningMe = false
	this.isReplyToMe = false

	this.author = data.user.screen_name
	this.authorId = data.user.id_str
	if(this.isRetweet) {
		this.originAuthor = data.retweeted_status.user.screen_name
		this.originAuthorId = data.retweeted_status.user.id_str
		this._text = data.retweeted_status.text
		this._entities = data.retweeted_status.entities
	} else {
		this.originAuthor = this.originAuthorId = null
		this._text = data.text
		this._entities = data.entities
	}
	this.replyTweet = {}
	if(this.isReply) {
		this.replyTweet.id = data.in_reply_to_status_id_str
		this.replyTweet.author = data.in_reply_to_screen_name
		this.replyTweet.authorId = data.in_reply_to_user_id_str
	}

	this._source = (data.source.substring(0,2) === '<a') ? cheerio.load(data.source)('a').text() : data.source
	this.source = this._source.replace(/ /g, '_')
	this._date = new Date(data.created_at)
	this.date = dateFormat(this._date, 'dd_mmm_HH:MM')
	this.mentions = []
	this.mentionIds = []
	if(this._entities.user_mentions !== undefined) {
		this._entities.user_mentions.forEach(function(user) {
			self.mentions.push(user.screen_name)
			self.mentionIds.push(user.id_str)
		})
	}
}
util.inherits(tweet, line)

tweet.prototype.filter = function(type, filters) {
	try {
		if(!util.isArray(filters))
			filters = [filters]

		if(filters.length <= 0) {
			return false
		}

		function filterCheck(filter, input) {
			if(filter.substring(0,1) === '/') {
				flags = filter.match(/\/([gim]*)$/)
				if(flags !== null)
					flags = flags[1]
				filter = new RegExp(filter, flags)
			} else {
				filter = filter.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&")
			}

			return (input.match(filter) !== null) ? true : false
		}

		var field = null
		if(type === 'user') {
			field = this.author
		} else if(type === 'retweeter') {
			field = this.originAuthor
		} else if(type === 'text') {
			field = this._text
		} else if(type === 'source') {
			field = this._source
		} else {
			return false // unknown filter type, fuck off
		}

		filters.forEach(function(filter) {
			if(!filterCheck(filter, field))
				throw new Error()
		})
	} catch(err) {
		return true
	}

	return false
}

tweet.prototype.build = function(width) {
	var self = this, wordwrap = require('./wordwrap')
	var indices = [], sliceOffset = 0, shiftOffset = 0, text = this._text, newText = '', offset = 0
	this.chunks = []

	// build up indices array, then we'll sort

	if(this.isRetweet)
		text = util.format('RT @%s ', this.originAuthor) + text
	text = this.author + ': ' + text
	indices.push([0, this.author.length, 'author', this.author, this.authorId])
	offset += this.author.length + 2
	if(this.isRetweet) {
		indices.push([offset + 3, offset + this.originAuthor.length + 4, 'mention', this.originAuthor, this.originAuthorId])
		offset += this.originAuthor.length + 5
	}

	if(this._entities.urls !== undefined) {
		this._entities.urls.forEach(function(url) {
			if(url.expanded_url.length <= 30) {
				url.indices[1] = url.indices[0] + url.expanded_url.length
				text.replace(url.url, url.expanded_url)
			}
			indices.push([url.indices[0] + offset, url.indices[1] + offset, 'url', url.url, url.expanded_url])
		})
	}
	if(this._entities.media !== undefined) {
		this._entities.media.forEach(function(url) {
			if(url.expanded_url.length <= 30) {
				url.indices[1] = url.indices[0] + url.expanded_url.length
				text.replace(url.url, url.expanded_url)
			}
			indices.push([url.indices[0] + offset, url.indices[1] + offset, 'url', url.url, url.expanded_url])
		})
	}
	if(this._entities.hashtags !== undefined) {
		this._entities.hashtags.forEach(function(tag) {
			indices.push([tag.indices[0] + offset, tag.indices[1] + offset, 'hashtag', tag.text])
		})
	}

	if(this._entities.user_mentions !== undefined) {
		this._entities.user_mentions.forEach(function(user) {
			indices.push([user.indices[0] + offset, user.indices[1] + offset, 'mention', user.screen_name, user.id_str])
		})
	}

	// insert date, via, source to end of text
	var offset = text.length, text = text + util.format(' (%s)', this.date)
	indices.push([offset, text.length, 'date', this.date, dateFormat(this._date, 'dd mmm HH:MM')])
	offset = text.length, text = text + ' via '
	indices.push([offset, text.length, 'extra'])
	offset = text.length, text = text + this.source
	indices.push([offset, text.length, 'source', this.source, this._source])

	// sort everything
	indices = _.sortBy(indices, function(entry) { return entry[0] })

	// A little post-processing to swap in expanded URLs
	indices = indices.map(function(entry) {
		entry[0] += shiftOffset, entry[1] += shiftOffset
		if(entry[2] === 'url' && entry[4].length <= 30) {
			shiftOffset += (entry[4].length - entry[3].length)
			text = text.replace(entry[3], entry[4])
		}
		return entry
	})

	// handle word wrap...forgive me for the following code, I know it's nasty.
	// we're looking for where the newlines were inserted basically, so we can modify the chunks appropriately
	newText = wordwrap(text, width - 2)
	var oldNewlineIndexes = [], insertIndexes = [], newLine = [], insert = [], inDiff = 0, re = /\n/g
	while((newLine = re.exec(text)) !== null) oldNewlineIndexes.push(newLine.index)
	while((newLine = re.exec(newText)) !== null) insertIndexes.push(newLine.index)
	oldNewlineIndexes.sort()
	insertIndexes.sort()
	inDiff = insertIndexes.length - oldNewlineIndexes.length
	if(inDiff !== 0) {
		for(var i = 0, j = 0, shiftOffset = 0; i < insertIndexes.length; i++) {
			if(insertIndexes[i] + shiftOffset !== oldNewlineIndexes[j]) {
				// we have a difference...we need to adjust and continue.
				shiftOffset++
				insert.push(insertIndexes[i])
			} else {
				j++
			}
		}
	}
	if(insert === null) insert = []

	var HTML_ENTITIES = {
		'&amp;': '&',
		'&gt;': '>',
		'&lt;': '<',
		'&quot;': '"',
		'&#39;': "'"
	}
	function cleanEntities(str) {
		return str.replace(/(&(?:amp|gt|lt|quot|#39);)/g, function(character) { return HTML_ENTITIES[character] })
	}

	// warning: below lies true madness.
	// this is a huge clusterfuck in order to adjust slices for word-wrapping,
	// yet still be able to trust twitter's slicing entities somewhat.
	var shiftOffset = 0, slice, from, to, insertCheck = []
	for(var i = 0; i < indices.length; i++) {
		insertCheck = insert.slice()
		if(indices[i][0] !== sliceOffset) {
			from = sliceOffset + shiftOffset
			to = indices[i][0] + shiftOffset

			// check for newlines where our slice should be, then adjust aim as needed
			insertCheck = insertCheck.reduce(function(cur, offset) {
				return cur + ((offset >= from && offset <= to) ? 1 : 0)
			}, 0)
			shiftOffset += insertCheck

			// gap between bits - interpret as text.
			this.chunks.push({
				type: 'text',
				content: cleanEntities(newText.slice(from, to + insertCheck)),
				from: from,
				to: to + insertCheck,
			})
			insertCheck = insert.slice()
		}

		from = indices[i][0] + shiftOffset
		to = indices[i][1] + shiftOffset
		// check for newlines where our slice should be, then adjust aim as needed
		insertCheck = insertCheck.reduce(function(cur, offset) {
			return cur + ((offset >= from && offset <= to) ? 1 : 0)
		}, 0)
		shiftOffset += insertCheck

		this.chunks.push({
			type: indices[i][2],
			content: cleanEntities(newText.slice(from, to + insertCheck)),
			from: from,
			to: to + insertCheck,
			extra: indices[i].slice(3)
		})
		sliceOffset = indices[i][1]
	}
	if(sliceOffset !== newText.length) {
		this.chunks.push({
			type: 'text',
			content: cleanEntities(newText.slice(sliceOffset + shiftOffset)),
			from: sliceOffset + shiftOffset,
			to: newText.length,
		})
	}

	this.width = width
	this.height = insertIndexes.length + 1

	return this
}

module.exports = tweet
