var EventEmitter = require('events').EventEmitter,
	util = require('util'),
	tweet = require('./tweet'),
	hook = require('./hook'),
	twit = require('twit')

// @todo merge with filter.js?

// Inherits eventemitter. Will handle filtering of tweets and the like as well.
function tweetstream(oauth, myId, filter) {
	EventEmitter.call(this)
	this.t = new twit(oauth)
	this.filter = filter
	this.myId = '' + myId // force string storage representation
	this.me = null
	this.friends = []
	this.stream = null
}
util.inherits(tweetstream, EventEmitter)

tweetstream.prototype.connect = function(fn) {
	var self = this

	this.t.get('users/show', { user_id: this.myId }, function(err, user) {
		if(err) {
			self.emit('twitter:error', err)
		} else {
			self.me = user
		}
	})

	this.stream = this.t.stream('user', { with: 'followings' })
	this.stream.on('friends', function(event) {
		self.friends.concat(event.friends)
	})
	this.stream.on('tweet', function(tweetData) {
		var tweetObj = new tweet(tweetData), filterTypes = ['user', 'text', 'source']
		if(tweetObj.isRetweet) filterTypes.concat('retweeter')

		// sift this tweet through our filters
		try {
			// @todo extended filter logic...at some point.
			filterTypes.forEach(function(type) {
				if(tweetObj.filter(type, self.filter.get(type)))
					throw new Error('FILTERED: ' + type) // bail out and return, we're ignoring this shit.
			})
		} catch(err) {
			if(err.message.match(/^FILTERED/) === null)
				throw err

			self.emit('stream:filteredtweet', tweetObj, err.message) // useful for debugging purposes
		}

		// post-process just a little, see if this is one of our tweets or what
		tweetObj.isMine = (tweetObj.authorId === self.myId) ? true : false
		tweetObj.isMentioningMe = (tweetObj.mentionIds.indexOf(self.myId) !== -1) ? true : false
		if(tweetObj.isReply)
			tweetObj.isReplyToMe = (tweetObj.replyTweet.authorId === self.myId) ? true : false

		// emit and then let scrollbuffer catch it, insert it, etc.
		self.emit('stream:tweet', tweetObj)
		//hook.emit('twitter:api:newtweet', tweetObj)
	})
	this.stream.on('delete', function(event) {
		// if only this was gonna be so simple...
		self.emit('stream:deletedtweet', event.status.id_str)
	})
	this.stream.on('error', function(err) {
		self.emit('stream:error', err)
	})
	this.stream.on('connect', function() {
		self.emit('stream:connect')
	})
	this.stream.on('reconnect', function() {
		//hook.emit('twitter:api:reconnect')
		self.emit.apply(['stream:reconnect'].concat(Array.prototype.slice.call(arguments)))
	})
	this.stream.on('disconnect', function() {
		//hook.emit('twitter:api:disconnect')
		self.emit.apply(['stream:disconnect'].concat(Array.prototype.slice.call(arguments)))
	})

	fn()
}

tweetstream.prototype.newTweet = function(message, replyId) {
	var self = this, params = { status: message }
	if(replyId)
		params.in_reply_to_status_id = replyId
	this.t.post('statuses/update', params, function(err, reply) {
		if(err) self.emit('twitter:error', err)
		// @todo what do with reply? o_O
	})
}

tweetstream.prototype.newRetweet = function(tweetid) {
	var self = this
	this.t.post('statuses/retweet/' + tweetid, {}, function(err, reply) {
		if(err) self.emit('twitter:error', err)
	})
}

tweetstream.prototype.favTweet = function(tweetid) {
	var self = this
	this.t.post('favorites/create', { id: tweetid }, function(err, reply) {
		if(err) self.emit('twitter:error', err)
	})
}

tweetstream.prototype.unfavTweet = function(tweetid) {
	var self = this
	this.t.post('favorites/destroy', { id: tweetid }, function(err, reply) {
		if(err) self.emit('twitter:error', err)
	})
}

tweetstream.prototype.deleteTweet = function(tweetid) {
	var self = this
	this.t.post('statuses/destroy/' + tweetid, {}, function(err, reply) {
		if(err) self.emit('twitter:error', err)
	})
}

tweetstream.prototype.getBackfill = function(count, fn) {
	var self = this
	this.t.get('statuses/home_timeline', {include_entities:true, count: count }, function(err, reply) {
		if(err) {
			self.emit('twitter:error', err)
			fn(err)
		} else {
			reply = reply.reverse().map(function(tweetData) {
				return new tweet(tweetData)
			})
			fn(null, reply)
		}
	})
}

tweetstream.prototype.getMentions = function(count, fn) {
	var self = this
	if(isNaN(count)) count = 20
	if(count > 100) count = 100
	this.t.get('statuses/mentions_timeline', {include_entities:true, count: count }, function(err, reply) {
		if(err) {
			self.emit('twitter:error', err)
			fn(err)
		} else {
			reply = reply.reverse().map(function(tweetData) {
				return new tweet(tweetData)
			})
			fn(null, reply)
		}
	})
}

tweetstream.prototype.close = function() {
	if(this.stream !== null)
		this.stream.stop()
}

module.exports = tweetstream
